#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A command-line tool which uses `pytwis` to interact with the Redis database of a Twitter toy clone.

To see the help information,

.. code:: bash

    $ ./pytwis_clt.py -h
    $ ./pytwis_clt.py --help

After launching `pytwis_clt.py`, you will be able to use the following commands:

* Register a new user:

.. code:: bash

    127.0.0.1:6379> register {username} {password} 

* Log into a user:

.. code:: bash

    127.0.0.1:6379> login {username} {password} 

* Log out of a user:

.. code:: bash

    127.0.0.1:6379> logout

* Change the password:

.. code:: bash

    127.0.0.1:6379> changepwd {old_password} {new_password} {confirmed_new_password}
    
* Get the profile of the current user:

.. code:: bash

    127.0.0.1:6379> userprofile

* Post a tweet:

.. code:: bash

    127.0.0.1:6379> post {tweet}

* Follow a user:

.. code:: bash

    127.0.0.1:6379> follow {followee_username}

* Unfollow a user:

.. code:: bash

    127.0.0.1:6379> unfollow {followee_username}

* Get the follower list:

.. code:: bash

    127.0.0.1:6379> followers

* Get the following list:

.. code:: bash

    127.0.0.1:6379> followings

* Get the timeline:

.. code:: bash

    127.0.0.1:6379> timeline
    127.0.0.1:6379> timeline {max_tweet_count}
    
Note that if a user is logged in, `timeline` will return the user timeline; 
otherwise `timeline` will return the general timeline.

* Exit the program:

.. code:: bash

    127.0.0.1:6379> exit
    127.0.0.1:6379> quit

"""

import argparse
import datetime
import parse
import sys

from pytwis import Pytwis


def validate_command(raw_command):
    """Validate the command input.
    
    Currently we only check the number of arguments according to the command type.
    
    Parameters
    ----------
    raw_command: str
        The raw command input, e.g., `register xxxxxx yyyyyy`.
        
    Raises
    ------
    ValueError
        If the raw command input doesn't have the correct number of arguments.
    """
    parsed_command = raw_command.split()
    arg_count = len(parsed_command) - 1

    if (len(parsed_command) == 0):
        return

    if (parsed_command[0] == 'register'):
        if (arg_count < 2):
            raise ValueError('register {username} {password}')
    elif (parsed_command[0] == 'login'):
        if (arg_count < 2):
            raise ValueError('login {username} {password}')
    elif (parsed_command[0] == 'logout'):
        pass
    elif (parsed_command[0] == 'changepwd'):
        if (arg_count < 3):
            raise ValueError('changepwd {old_password} {new_password} {confirmed_new_password}')
    elif (parsed_command[0] == 'userprofile'):
        pass
    elif (parsed_command[0] == 'post'):
        if (arg_count < 1):
            raise ValueError('post {tweet}')
    elif (parsed_command[0] == 'follow'):
        if (arg_count < 1):
            raise ValueError('follow {followee_username}')
    elif (parsed_command[0] == 'unfollow'):
        if (arg_count < 1):
            raise ValueError('unfollow {followee_username}')
    elif (parsed_command[0] == 'followers'):
        pass
    elif (parsed_command[0] == 'followings'):
        pass
    elif (parsed_command[0] == 'timeline'):
        if (arg_count > 2):
            raise ValueError('timeline {max_tweet count} or timeline')
    elif (parsed_command[0] == 'exit') or (parsed_command[0] == 'quit'):
        pass
    else:
        raise ValueError('Invalid pytwis command')


def pytwis_command_parser(raw_command):
    """Parse the command input.
    
    Parameters
    ----------
    raw_command: str
        The raw command input, e.g., `register xxxxxx yyyyyy`.
        
    Returns
    -------
    command_with_args: list(str, dict(str, str or int))
        The parsed command output. The first element of the list is the command type, e.g., 'register', 
        and the second element is the command arguments, e.g., {'username': <username>, 'password': <password>} 
        for `register`.
    
    Raises
    ------
    ValueError
        If the raw command can't be parsed correctly, e.g., it has an incorrect number of arguments or 
        incorrect arguments.
    """
    # Separate the command from its arguments.
    splited_raw_command = raw_command.split(' ', 1)
    command_with_args = [splited_raw_command[0]]

    # Some command (e.g., logout) may not have arguments.
    arg_dict = {}

    validate_command(raw_command)

    if command_with_args[0] == 'register':
        # register must have two arguments: username and password.
        args = splited_raw_command[1]
        arg_dict = parse.parse('{username} {password}', args)
        if arg_dict is None:
            raise ValueError('register has incorrect arguments')
        elif ' ' in arg_dict['password']:
            raise ValueError("password can't contain spaces")

        print('register: username = {}, password = {}'.format(arg_dict['username'], arg_dict['password']))
    elif command_with_args[0] == 'login':
        # login must have two arguments: username and password.
        args = splited_raw_command[1]
        arg_dict = parse.parse('{username} {password}', args)
        if arg_dict is None:
            raise ValueError('login has incorrect arguments')

        print('login: username = {}, password = {}'.format(arg_dict['username'], arg_dict['password']))
    elif command_with_args[0] == 'logout':
        # logout doesn't have any arguments.
        pass
    elif command_with_args[0] == 'changepwd':
        # changepwd must have three arguments: old_password, new_password, and confirmed_new_password.
        args = splited_raw_command[1]
        arg_dict = parse.parse('{old_password} {new_password} {confirmed_new_password}', args)
        if arg_dict is None:
            raise ValueError('changepwd has incorrect arguments')
        elif arg_dict['new_password'] != arg_dict['confirmed_new_password']:
            raise ValueError('The confirmed new password is different from the new password')
        elif arg_dict['new_password'] == arg_dict['old_password']:
            raise ValueError('The new password is the same as the old password')

        print('changepwd: old = {}, new = {}'.format(arg_dict['old_password'], arg_dict['new_password']))
    elif command_with_args[0] == 'userprofile':
        # userprofile doesn't have any arguments.
        pass
    elif command_with_args[0] == 'post':
        # post must have one argument: tweet
        arg_dict = {'tweet': splited_raw_command[1]}
    elif command_with_args[0] == 'follow':
        # follow must have one argument: followee.
        arg_dict = {'followee': splited_raw_command[1]}
    elif command_with_args[0] == 'unfollow':
        # unfollow must have one argument: followee.
        arg_dict = {'followee': splited_raw_command[1]}
    elif command_with_args[0] == 'followers':
        # followers doesn't have any arguments.
        pass
    elif command_with_args[0] == 'followings':
        # followings doesn't have any arguments.
        pass
    elif command_with_args[0] == 'timeline':
        # timeline has either zero or one argument.
        max_cnt_tweets = -1
        if len(splited_raw_command) >= 2:
            max_cnt_tweets = int(splited_raw_command[1])

        arg_dict = {'max_cnt_tweets': max_cnt_tweets}
    elif command_with_args[0] == 'exit' or command_with_args[0] == 'quit':
        # exit or quit doesn't have any arguments.
        pass
    else:
        pass

    command_with_args.append(arg_dict)

    return command_with_args


def print_tweets(tweets):
    """Print a list of tweets one by one separated by "="s.
    
    Parameters
    ----------
    tweets: list(dict)
        A list of tweets. Each tweet is a dict containing the username of the tweet's author, the post time, 
        and the tweet body.
    """
    print('=' * 60)
    for index, tweet in enumerate(tweets):
        print('-' * 60)
        print('Tweet {}:'.format(index))
        print('Username:', tweet['username'])
        print('Time:', datetime.datetime.fromtimestamp(int(tweet['unix_time'])).strftime('%Y-%m-%d %H:%M:%S'))
        print('Body:\n\t', tweet['body'])
        print('-' * 60)
    print('=' * 60)


def pytwis_command_processor(twis, auth_secret, command_with_args):
    """Process the parsed command.
    
    Parameters
    ----------
    twis: Pytwis
        A Pytwis instance which interacts with the Redis database of the Twitter toy clone.
    auth_secret: str
        The authentication secret of a logged-in user.
    command_with_args:
        The parsed command output by pytwis_command_parser().
    """
    command = command_with_args[0]
    args = command_with_args[1]

    if command == 'register':
        succeeded, result = twis.register(args['username'], args['password'])
        if succeeded:
            print('Registered {}'.format(args['username']))
        else:
            print("Couldn't register {} with error = {}".format(args['username'], result['error']))
    elif command == 'login':
        succeeded, result = twis.login(args['username'], args['password'])
        if succeeded:
            auth_secret[0] = result['auth']
            print('Logged into username {}'.format(args['username']))
        else:
            print("Couldn't log into username {} with error = {}".format(args['username'], result['error']))
    elif command == 'logout':
        succeeded, result = twis.logout(auth_secret[0])
        if succeeded:
            auth_secret[0] = result['auth']
            print('Logged out of username {}'.format(result['username']))
        else:
            print("Couldn't log out with error = {}".format(result['error']))
    elif command == 'changepwd':
        succeeded, result = twis.change_password(auth_secret[0], args['old_password'], args['new_password'])
        if succeeded:
            auth_secret[0] = result['auth']
            print('Changed the password')
        else:
            print("Couldn't change the password with error = {}".format(result['error']))
    elif command == 'userprofile':
        succeeded, result = twis.get_user_profile(auth_secret[0])
        if succeeded:
            print('Got the user profile')
            print('=' * 20)
            for key, value in result.items():
                print('{}: {}'.format(key, value))
            print('=' * 20)
        else:
            print("Couldn't get the user profile with error = {}".format(result['error']))
    elif command == 'post':
        succeeded, result = twis.post_tweet(auth_secret[0], args['tweet'])
        if succeeded:
            print('Posted the tweet')
        else:
            print("Couldn't post the tweet with error = {}".format(result['error']))
    elif command == 'follow':
        succeeded, result = twis.follow(auth_secret[0], args['followee'])
        if succeeded:
            print('Followed username {}'.format(args['followee']))
        else:
            print("Couldn't follow the username {} with error = {}".format(args['followee'], result['error']))
    elif command == 'unfollow':
        succeeded, result = twis.unfollow(auth_secret[0], args['followee'])
        if succeeded:
            print('Unfollowed username {}'.format(args['followee']))
        else:
            print("Couldn't unfollow the username {} with error = {}".format(args['followee'], result['error']))
    elif command == 'followers':
        succeeded, result = twis.get_followers(auth_secret[0])
        if succeeded:
            print('Got the list of {} followers'.format(len(result['follower_list'])))
            print('=' * 20)
            for follower in result['follower_list']:
                print('\t' + follower)
            print('=' * 20)
        else:
            print("Couldn't get the follower list with error = {}".format(result['error']))
    elif command == 'followings':
        succeeded, result = twis.get_following(auth_secret[0])
        if succeeded:
            print('Got the list of {} followings'.format(len(result['following_list'])))
            print('=' * 60)
            for following in result['following_list']:
                print('\t' + following)
            print('=' * 60)
        else:
            print("Couldn't get the following list with error = {}".format(result['error']))
    elif command == 'timeline':
        succeeded, result = twis.get_timeline(auth_secret[0], args['max_cnt_tweets'])
        if succeeded:
            if auth_secret[0] != '':
                print('Got {} tweets in the user timeline'.format(len(result['tweets'])))
            else:
                print('Got {} tweets in the general timeline'.format(len(result['tweets'])))
            print_tweets(result['tweets'])
        else:
            if auth_secret[0] != '':
                print("Couldn't get the user timeline with error = {}".format(result['error']))
            else:
                print("Couldn't get the general timeline with error = {}".format(result['error']))
    else:
        pass


def pytwis_cli():
    """The main routine of this command-line tool."""
    # Note that we set the conflict handler of ArgumentParser to 'resolve' because we reuse the short help 
    # option '-h' for the host name.
    parser = argparse.ArgumentParser(conflict_handler="resolve", description=\
                                         'Connect to the Redis database of a Twitter clone and '
                                         'then run commands to access and update the database.')
    # TODO: Add epilog for the help information about online commands after connecting to the Twitter clone.
    parser.add_argument('-h', '--hostname', nargs='?', default='127.0.0.1',
                        help='''the Redis server hostname. If the option is not specified, will be defaulted to 127.0.0.1. 
                             If the option is specified but no value is given after the option, then the help information 
                             is displayed instead.
                             ''')
    parser.add_argument('-p', '--port', default=6379,
                        help='the Redis server port. If the option is not specified, will be defaulted to 6379.')
    parser.add_argument('-n', '--db', default=0,
                        help='the Redis server database. If the option is not specified, will be defaulted to 0.')
    parser.add_argument('-a', '--password', default='',
                        help='the Redis server password. If the option not specified, will be defaulted to an empty string.')

    args = parser.parse_args()

    # If no value is given after the option '-h', then the help information is displayed.
    if args.hostname is None:
        parser.print_help()
        return 0

    print('The input Redis server hostname is {}.'.format(args.hostname))
    print('The input Redis server port is {}.'.format(args.port))
    print('The input Redis server database is {}.'.format(args.db))
    if args.password != '':
        print('The input Redis server password is "{}".'.format(args.password))
    else:
        print('The input Redis server password is empty.')

    try:
        twis = Pytwis(args.hostname, args.port, args.db, args.password)
    except ValueError as e:
        print('Failed to connect to the Redis server: {}'.format(str(e)),
              file=sys.stderr)
        return -1

    auth_secret = ['']
    while True:
        try:
            command_with_args = pytwis_command_parser(
                input('Please enter a command '
                      '(register, login, logout, changepwd, userprofile, post, '
                      'follow, unfollow, followers, followings, timeline):\n{}:{}> ' \
                      .format(args.hostname, args.port)))
            if command_with_args[0] == "exit" or command_with_args[0] == 'quit':
                # Log out of the current user before exiting.
                if len(auth_secret[0]) > 0:
                    pytwis_command_processor(twis, auth_secret, ['logout', {}])
                print('pytwis is exiting.')
                return 0;

        except ValueError as e:
            print('Invalid pytwis command: {}'.format(str(e)),
                  file=sys.stderr)
            continue

        pytwis_command_processor(twis, auth_secret, command_with_args)


if __name__ == "__main__":
    pytwis_cli()



# -*- coding: utf-8 -*-
"""pytwis -- A Twitter-toy-clone backend using Python and Redis.

This module implements the backend for a simplified Twitter clone based on Redis. 
We follow the Redis tutorial (https://redis.io/topics/twitter-clone) to design 
the data layout of the Redis database.  

It supports the following features:

-  Register new users
-  Log in/out
-  Change user password
-  Post tweets
-  Follower/Following
-  General timeline for anonymous user
-  User timeline

TODOs:

-  #hashtags
-  @mentions
-  Retweets
-  Replies
-  Conversations
-  Edit/Delete tweets
-  And more

"""

import redis
from redis.exceptions import (ResponseError, TimeoutError, WatchError)
import secrets
import time

class Pytwis:
    """This class implements all the interfaces to the Redis database of the Twitter-toy-clone."""
    
    REDIS_SOCKET_CONNECT_TIMEOUT = 60
    
    NEXT_USER_ID_KEY = 'next_user_id'
    
    USERS_HASH_KEY = 'users'
    
    USER_ID_PROFILE_KEY_FORMAT = 'user:{}'
    USER_ID_PROFILE_USERNAME_KEY = 'username'
    USER_ID_PROFILE_PASSWORD_KEY = 'password'
    USER_ID_PROFILE_AUTH_KEY = 'auth'
    
    AUTHS_HASH_KEY = 'auths'
    
    FOLLOWER_ZSET_KEY_FORMAT = 'follower:{}'
    FOLLOWING_ZSET_KEY_FORMAT = 'following:{}'
    
    NEXT_POST_ID_KEY = 'next_post_id'
    
    POST_ID_KEY_FORMAT = 'post:{}'
    POST_ID_USERID_KEY = 'userid'
    POST_ID_UNIXTIME_KEY = 'unix_time'
    POST_ID_BODY_KEY = 'body'
    
    GENERAL_TIMELINE_KEY = 'timeline'
    GENERAL_TIMELINE_MAX_POST_CNT = 1000
    
    POST_ID_USER_KEY_FORMAT = 'posts:{}'
    
    def __init__(self, hostname='127.0.0.1', port=6379, db=0, password =''):
        """Initialize the class Pytiws.
        
        Parameters
        ----------
        hostname : str, optional
            The Redis server hostname which is usually an IP address (default: 127.0.0.1).
        port : int, opti
            The Redis server port number (default: 6379).
        db : int 
            The selected Redis database index (default: 0).
        password : str)
            The Redis server password (default: '').
                
        Raises
        ------
        ValueError
            If failed to connect to the Redis server with either ResponseError or TimeoutError.
        """
        # TODO: Set unix_socket_path='/tmp/redis.sock' to use Unix domain socket 
        # if the host name is 'localhost'. Note that need to uncomment the following 
        # line in /etc/redis/redis.conf:
        #
        # unixsocket /tmp/redis.sock
        # 
        self._rc = redis.StrictRedis(
            host=hostname,
            port=port,
            db=db,
            password=password,
            decode_responses=True, # Decode the response bytes into strings.
            socket_connect_timeout=self.REDIS_SOCKET_CONNECT_TIMEOUT)
        
        # Test the connection by ping.
        try:
            if self._rc.ping() == True:
                print('Ping {} returned True'.format(hostname))
        except (ResponseError, TimeoutError) as e:
            raise ValueError(str(e)) from e
        
    def _is_loggedin(self, auth_secret):
        """Check if a user is logged-in by verifying the input authentication secret.
        
        Parameters
        ----------
        auth_secret: str
            The authentication secret of a logged-in user.
            
        Returns
        -------
        bool 
            True if the authentication secret is valid, False otherwise.
        userid: str
            The user ID associated with the authentication secret if the authentication secret 
            valid, None otherwise. 
        """
        # Get the userid from the authentication secret.
        userid = self._rc.hget(self.AUTHS_HASH_KEY, auth_secret)
        if userid is None:
            return (False, None)
        
        # Compare the input authentication secret with the stored one.
        userid_profile_key = self.USER_ID_PROFILE_KEY_FORMAT.format(userid)
        stored_auth_secret = self._rc.hget(userid_profile_key, self.USER_ID_PROFILE_AUTH_KEY)
        if auth_secret == stored_auth_secret:
            return (True, userid)
        else:
            # TODO: Resolve the inconsistency of the two authentication secrets. 
            return (False, None)

    def register(self, username, password):
        """Register a new user.
        
        Parameters
        ----------
        username: str
            The username.
        password: str
            The password.
            
        Returns
        -------
        bool
            True if the new user is successfully registered, False otherwise.
        result
            An empty dict if the new user is successfully registered, a dict 
            containing the error string with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'username {} already exists'.format(username)
        """
        result = {'error': None}
        
        # TODO: add the username check.
        # TODO: add the password check.
        # https://stackoverflow.com/questions/16709638/checking-the-strength-of-a-password-how-to-check-conditions
        
        # Update the username-to-userid mapping.
        with self._rc.pipeline() as pipe:
            while True:
                try:
                    # Put a watch on the Hash 'users': username -> user-id, in case that 
                    # multiple clients are registering with the same username.
                    pipe.watch(self.USERS_HASH_KEY)
                    username_exists = pipe.hexists(self.USERS_HASH_KEY, username)
                    if username_exists:
                        result['error'] = 'username {} already exists'.format(username)
                        return (False, result);
                    
                    # Get the next user-id. If the key "next_user_id" doesn't exist,
                    # it will be created and initialized as 0, and then incremented by 1.
                    userid = pipe.incr(self.NEXT_USER_ID_KEY)
                    
                    # Set the username-to-userid pair in USERS_HASH_KEY.
                    pipe.multi()
                    pipe.hset(self.USERS_HASH_KEY, username, userid)
                    pipe.execute()
                    
                    break
                except WatchError:
                    continue
                
            # Generate the authentication secret.
            auth_secret = secrets.token_hex()
            userid_profile_key = self.USER_ID_PROFILE_KEY_FORMAT.format(userid)
            
            pipe.multi()
            # Update the authentication_secret-to-userid mapping.
            pipe.hset(self.AUTHS_HASH_KEY, auth_secret, userid)
            # Create the user profile.
            # TODO: Store the hashed password instead of the raw password.
            pipe.hmset(userid_profile_key, 
                       {self.USER_ID_PROFILE_USERNAME_KEY: username, 
                        self.USER_ID_PROFILE_PASSWORD_KEY: password,
                        self.USER_ID_PROFILE_AUTH_KEY: auth_secret})
            pipe.execute()
        
        return (True, result)
            
    def change_password(self, auth_secret, old_password, new_password):
        """Change the user password.
        
        Parameters
        ----------
        auth_secret: str
            The authentication secret which will be used for user authentication.
        old_password: str
            The old password before the change.
        new_password: str
            The new password after the change.
        
        Returns
        -------
        bool
            True if the password is successfully changed, False otherwise.
        result
            A dict containing the new authentication secret with the key 'auth' 
            if the password is successfully changed, a dict containing the error 
            string with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'Not logged in'
        -  'Incorrect old password'
        """
        result = {'error': None}
        
        # Check if the user is logged in.
        loggedin, userid = self._is_loggedin(auth_secret)
        if not loggedin:
            result['error'] = 'Not logged in'
            return (False, result)
        
        # Check if the old password matches.
        userid_profile_key = self.USER_ID_PROFILE_KEY_FORMAT.format(userid)
        stored_password = self._rc.hget(userid_profile_key, self.USER_ID_PROFILE_PASSWORD_KEY)
        if stored_password != old_password:
            result['error'] = 'Incorrect old password'
            return (False, result)
        
        # TODO: add the new password check.
        
        # Generate the new authentication secret.
        new_auth_secret = secrets.token_hex()
        
        # Replace the old password by the new one and the old authentication secret by the new one.
        with self._rc.pipeline() as pipe:
            pipe.multi()
            pipe.hset(userid_profile_key, self.USER_ID_PROFILE_PASSWORD_KEY, new_password)
            pipe.hset(userid_profile_key, self.USER_ID_PROFILE_AUTH_KEY, new_auth_secret)
            pipe.hset(self.AUTHS_HASH_KEY, new_auth_secret, userid)
            pipe.hdel(self.AUTHS_HASH_KEY, auth_secret)
            pipe.execute()
        
        result[self.USER_ID_PROFILE_AUTH_KEY] = new_auth_secret
        return (True, result)
    
    def login(self, username, password):
        """Log into a user.
        
        Parameters
        ----------
        username: str
            The username.
        password: str
            The password.
        
        Returns
        -------
        bool
            True if the login is successful, False otherwise.
        result
            A dict containing the authentication secret with the key 'auth' 
            if the login is successful, a dict containing the error string 
            with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  "username {} doesn't exist".format(username)
        -  'Incorrect password'
        """
        result = {'error': None}
        
        # Get the user-id based on the username.
        userid = self._rc.hget(self.USERS_HASH_KEY, username)
        if userid is None:
            result['error'] = "username {} doesn't exist".format(username)
            return (False, result)
        
        # Compare the input password with the stored one. If it matches, 
        # return the authentication secret.
        userid_profile_key = self.USER_ID_PROFILE_KEY_FORMAT.format(userid)
        stored_password = self._rc.hget(userid_profile_key, self.USER_ID_PROFILE_PASSWORD_KEY)
        if password == stored_password:
            result[self.USER_ID_PROFILE_AUTH_KEY] = self._rc.hget(userid_profile_key, self.USER_ID_PROFILE_AUTH_KEY)
            return (True, result)
        else:
            result['error'] = 'Incorrect password'
            return (False, result)
    
    def logout(self, auth_secret):
        """Log out of a user.
        
        Parameters
        ----------
        auth_secret: str
            The authentication secret of the logged-in user.
        
        Returns
        -------
        bool
            True if the logout is successful, False otherwise.
        result
            None if the logout is successful, a dict containing the error string 
            with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'Not logged in'
        """
        result = {'error': None}
        
        # Check if the user is logged in.
        loggedin, userid = self._is_loggedin(auth_secret)
        if not loggedin:
            result['error'] = 'Not logged in'
            return (False, result)
        
        # Generate the new authentication secret.
        new_auth_secret = secrets.token_hex()
        
        # Replace the old authentication secret by the new one.
        userid_profile_key = self.USER_ID_PROFILE_KEY_FORMAT.format(userid)
        with self._rc.pipeline() as pipe:
            pipe.multi()
            pipe.hset(userid_profile_key, self.USER_ID_PROFILE_AUTH_KEY, new_auth_secret)
            pipe.hset(self.AUTHS_HASH_KEY, new_auth_secret, userid)
            pipe.hdel(self.AUTHS_HASH_KEY, auth_secret)
            pipe.execute()
            
        result[self.USER_ID_PROFILE_USERNAME_KEY] = self._rc.hget(userid_profile_key, self.USER_ID_PROFILE_USERNAME_KEY)
        result[self.USER_ID_PROFILE_AUTH_KEY] = ''
        return (True, result)
    
    def get_user_profile(self, auth_secret):
        """Get the profile (i.e., username, password, etc.) of a user.
        
        Parameters
        ----------
        auth_secret: str
            The authentication secret of the logged-in user.
        
        Returns
        -------
        bool
            True if the logout is successful, False otherwise.
        result
            A dict containing the following keys:
            
            -  USER_ID_PROFILE_USERNAME_KEY
            -  USER_ID_PROFILE_PASSWORD_KEY
            -  USER_ID_PROFILE_AUTH_KEY = 'auth'
            
            if the user profile is obtained successfully; otherwise a dict containing the error string 
            with the key 'error'.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'Not logged in'
        """
        result = {'error': None}
        
        # Check if the user is logged in.
        loggedin, userid = self._is_loggedin(auth_secret)
        if not loggedin:
            result['error'] = 'Not logged in'
            return (False, result)
        
        userid_profile_key = self.USER_ID_PROFILE_KEY_FORMAT.format(userid)
        result = self._rc.hgetall(userid_profile_key)
        
        return (True, result)
    
    def post_tweet(self, auth_secret, tweet):
        """Post a tweet.
        
        Parameters
        ----------
        auth_secret: str
            The authentication secret of the logged-in user.
        tweet: str
            The tweet that will be posted.
        
        Returns
        -------
        bool
            True if the tweet is successfully posted, False otherwise.
        result
            None if the tweet is successfully posted, a dict containing 
            the error string with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'Not logged in'
        """
        result = {'error': None}
        
        # Check if the user is logged in.
        loggedin, userid = self._is_loggedin(auth_secret)
        if not loggedin:
            result['error'] = 'Not logged in'
            return (False, result)
        
        # Get the next user-id. If the key "next_user_id" doesn't exist,
        # it will be created and initialized as 0, and then incremented by 1.
        post_id = self._rc.incr(self.NEXT_POST_ID_KEY)
        post_id_key = self.POST_ID_KEY_FORMAT.format(post_id)
        
        post_id_user_key = self.POST_ID_USER_KEY_FORMAT.format(userid)
        
        follower_zset_key = self.FOLLOWER_ZSET_KEY_FORMAT.format(userid)
        followers = self._rc.zrange(follower_zset_key, 0, -1)
        
        unix_time = int(time.time())
        with self._rc.pipeline() as pipe:
            pipe.multi()
            # Store the tweet with its user ID and UNIX timestamp.
            pipe.hmset(post_id_key,
                       {self.POST_ID_USERID_KEY: userid,
                        self.POST_ID_UNIXTIME_KEY: unix_time,
                        self.POST_ID_BODY_KEY: tweet})
            
            # Add the tweet to the user timeline.
            pipe.lpush(post_id_user_key, post_id)
            
            # Write fanout the tweet to all the followers' timelines.
            for follower in followers:
                post_id_follower_key = self.POST_ID_USER_KEY_FORMAT.format(follower)
                pipe.lpush(post_id_follower_key, post_id)
            
            # Add the tweet to the general timeline and left trim the general timeline to only retain 
            # the latest GENERAL_TIMELINE_MAX_POST_CNT tweets.
            pipe.lpush(self.GENERAL_TIMELINE_KEY, post_id)
            pipe.ltrim(self.GENERAL_TIMELINE_KEY, 0, self.GENERAL_TIMELINE_MAX_POST_CNT - 1)
            
            pipe.execute()
        
        return (True, result)
    
    def follow(self, auth_secret, followee_username):
        """Follow a user.
        
        Parameters
        ----------
        auth_secret: str
            The authentication secret of the logged-in user.
        followee_username: str
            The username of the followee.
        
        Returns
        -------
        bool
            True if the follow is successful, False otherwise.
        result
            None if the follow is successful, a dict containing 
            the error string with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'Not logged in'
        -  "Followee {} doesn't exist".format(followee_username)
        -  "Can't follow yourself {}".format(followee_username)
        """
        result = {'error': None}
        
        # Check if the user is logged in.
        loggedin, userid = self._is_loggedin(auth_secret)
        if not loggedin:
            result['error'] = 'Not logged in'
            return (False, result)
        
        with self._rc.pipeline() as pipe:
            # Check if the followee exists.
            while True:
                try:
                    # Put a watch on the Hash 'users': username -> user-id, in case that 
                    # other clients are modifying the Hash 'users'.
                    pipe.watch(self.USERS_HASH_KEY)
                    followee_userid = pipe.hget(self.USERS_HASH_KEY, followee_username)
                    if followee_userid is None:
                        result['error'] = "Followee {} doesn't exist".format(followee_username)
                        return (False, result);
                    elif followee_userid == userid:
                        result['error'] = "Can't follow yourself {}".format(followee_username)
                        return (False, result)
                    
                    break
                except WatchError:
                    continue
            
            # Update the two zset 'followers:[followee_username]' and 'following:[username]'.
            follower_zset_key = self.FOLLOWER_ZSET_KEY_FORMAT.format(followee_userid)
            following_zset_key = self.FOLLOWING_ZSET_KEY_FORMAT.format(userid)
            unix_time = int(time.time())
            pipe.multi()
            pipe.zadd(follower_zset_key, unix_time, userid)
            pipe.zadd(following_zset_key, unix_time, followee_userid)
            pipe.execute()
            
        return (True, result)
    
    def unfollow(self, auth_secret, followee_username):
        """Unfollow a user.
        
        Parameters
        ----------
        auth_secret: str
            The authentication secret of the logged-in user.
        followee_username: str
            The username of the followee.
        
        Returns
        -------
        bool
            True if the unfollow is successful, False otherwise.
        result
            None if the unfollow is successful, a dict containing 
            the error string with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'Not logged in'
        -  "Followee {} doesn't exist".format(followee_username)
        """
        result = {'error': None}
        
        # Check if the user is logged in.
        loggedin, userid = self._is_loggedin(auth_secret)
        if not loggedin:
            result['error'] = 'Not logged in'
            return (False, result)
        
        with self._rc.pipeline() as pipe:
            # Check if the followee exists.
            while True:
                try:
                    # Put a watch on the Hash 'users': username -> user-id, in case that 
                    # other clients are modifying the Hash 'users'.
                    pipe.watch(self.USERS_HASH_KEY)
                    followee_userid = pipe.hget(self.USERS_HASH_KEY, followee_username)
                    if followee_userid is None:
                        result['error'] = "Followee {} doesn't exist".format(followee_username)
                        return (False, result);
                    
                    break
                except WatchError:
                    continue
            
            # Remove followee_userid from the zset 'following:[username]' and remove userid 
            # from the zset 'followers:[followee_username]'.
            follower_zset_key = self.FOLLOWER_ZSET_KEY_FORMAT.format(followee_userid)
            following_zset_key = self.FOLLOWING_ZSET_KEY_FORMAT.format(userid)
            pipe.multi()
            pipe.zrem(follower_zset_key, userid)
            pipe.zrem(following_zset_key, followee_userid)
            pipe.execute()
            
        return (True, result)
    
    def get_followers(self, auth_secret):
        """Get the follower list of a logged-in user.
        
        Parameters
        ----------
        auth_secret: str
            The authentication secret of the logged-in user.
        
        Returns
        -------
        bool
            True if the follower list is successfully obtained, False otherwise.
        result
            A dict containing the follower list with the key 'follower_list' 
            if the follower list is successfully obtained, a dict containing 
            the error string with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'Not logged in'
        """
        result = {'error': None}
        
        # Check if the user is logged in.
        loggedin, userid = self._is_loggedin(auth_secret)
        if not loggedin:
            result['error'] = 'Not logged in'
            return (False, result)
        
        # Get the list of followers' userids.
        follower_zset_key = self.FOLLOWER_ZSET_KEY_FORMAT.format(userid)
        follower_userids = self._rc.zrange(follower_zset_key, 0, -1)
        
        if follower_userids is None or len(follower_userids) == 0:
            result['follower_list'] = []
            return (True, result)
        
        # Get the list of followers' usernames from their userids.
        with self._rc.pipeline() as pipe:
            pipe.multi()
            
            for follower_userid in follower_userids:
                follower_userid_profile_key = self.USER_ID_PROFILE_KEY_FORMAT.format(follower_userid)
                pipe.hget(follower_userid_profile_key, self.USER_ID_PROFILE_USERNAME_KEY)
            
            result['follower_list'] = pipe.execute()
            
        return (True, result)
        
    def get_following(self, auth_secret):
        """Get the following list of a logged-in user.
        
        Parameters
        ----------
        auth_secret: str
            The authentication secret of the logged-in user.
        
        Returns
        -------
        bool
            True if the following list is successfully obtained, False otherwise.
        result
            A dict containing the following list with the key 'following_list' 
            if the follower list is successfully obtained, a dict containing 
            the error string with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'Not logged in'
        """
        result = {'error': None}
        
        # Check if the user is logged in.
        loggedin, userid = self._is_loggedin(auth_secret)
        if not loggedin:
            result['error'] = 'Not logged in'
            return (False, result)
        
        # Get the list of followers' userids.
        following_zset_key = self.FOLLOWING_ZSET_KEY_FORMAT.format(userid)
        following_userids = self._rc.zrange(following_zset_key, 0, -1)
        
        if following_userids is None or len(following_userids) == 0:
            result['following_list'] = []
            return (True, result)
        
        # Get the list of followings' usernames from their userids.
        with self._rc.pipeline() as pipe:
            pipe.multi()
            
            for following_userid in following_userids:
                following_userid_profile_key = self.USER_ID_PROFILE_KEY_FORMAT.format(following_userid)
                pipe.hget(following_userid_profile_key, self.USER_ID_PROFILE_USERNAME_KEY)
            
            result['following_list'] = pipe.execute()
            
        return (True, result)
    
    def get_timeline(self, auth_secret, max_cnt_tweets):
        """Get the general or user timeline. 
        
        If an empty authentication secret is given, this method will return the general timeline. 
        If an authentication secret is given and it is valid, this method will return the user timeline.
        If an authentication secret is given but it is invalid, this method will return an error. 
        
        Parameters
        ----------
        auth_secret: str
            Either the authentication secret of the logged-in user or an empty string.
        max_cnt_tweets: int
            The maximum number of tweets included in the timeline. If it is set to -1, 
            then all the available tweets will be included.
        
        Returns
        -------
        bool
            True if the timeline is successfully retrieved, False otherwise.
        result
            A dict containing a list of tweets with the key 'tweets' if 
            the timeline is successfully retrieved, a dict containing 
            the error string with the key 'error' otherwise.
            
        Note
        ----
        Possible error strings are listed as below: 
        
        -  'Not logged in'
        """
        result = {'error': None}
        
        if auth_secret == '':
            # An empty authentication secret implies getting the general timeline.
            timeline_key = self.GENERAL_TIMELINE_KEY
        else:
            # Check if the user is logged in.
            loggedin, userid = self._is_loggedin(auth_secret)
            if not loggedin:
                result['error'] = 'Not logged in'
                return (False, result)
            
            # Get the user timeline.
            timeline_key = self.POST_ID_USER_KEY_FORMAT.format(userid)
        
        result['tweets'] = []
        if max_cnt_tweets == 0:
            return (True, result)
        elif max_cnt_tweets == -1:
            # Return all the tweets in the timeline.
            last_tweet_index = -1
        else:
            # Return at most max_cnt_tweets tweets.
            last_tweet_index = max_cnt_tweets - 1
            
        # Get the post IDs of the tweets.
        post_ids = self._rc.lrange(timeline_key, 0, last_tweet_index)
        
        if len(post_ids) == 0:
            return (True, result)
        
        with self._rc.pipeline() as pipe:
            # Get the tweets with their user IDs and UNIX timestamps.
            pipe.multi()
            for post_id in post_ids:
                post_id_key = self.POST_ID_KEY_FORMAT.format(post_id)
                pipe.hgetall(post_id_key)
            result['tweets'] = pipe.execute()
        
            # Get the userid-to-username mappings for all the user IDs associated with the tweets.
            userid_set = { tweet[self.POST_ID_USERID_KEY] for tweet in result['tweets'] }
            userid_list = []
            pipe.multi()
            for userid in userid_set:
                userid_list.append(userid)
                userid_key = self.USER_ID_PROFILE_KEY_FORMAT.format(userid)
                pipe.hget(userid_key, self.USER_ID_PROFILE_USERNAME_KEY)
            username_list = pipe.execute()
        
        userid_to_username = { userid: username for userid, username in zip(userid_list, username_list) }
        
        # Add the username for the user ID of each tweet.
        for tweet in result['tweets']:
            tweet[self.USER_ID_PROFILE_USERNAME_KEY] = userid_to_username[tweet[self.POST_ID_USERID_KEY]]
        
        return (True, result)